from typing import Any, Mapping
from warnings import warn

from sqlalchemy import func

from .dates import CreatedAtMixin, UpdatedAtMixin, CreatedUpdatedAtMixin
from .states import EnabledMixin, EnabledNoMixin

__version__ = "0.1.2"
__all__ = [
    'CreatedAtMixin', 'UpdatedAtMixin', 'CreatedUpdatedAtMixin',
    'EnabledMixin', 'EnabledNoMixin',
    'UpdateFromDictMixin',
]


class UpdateFromDictMixin:
    def update_from_dict(self, updates: Mapping[str, Any], *, always_set_last_update=False):
        """
        Update the model in-place from a dictionary of attributes.
        If the model has a column named ``<attr>_last_update`` and ``<attr>`` is in the updates dictionary,
        then ``<attr>_last_update`` is set to the current time if the new value is different,
        or if ``always_set_last_update`` is set to ``true``.

        :param updates: dictionary of attributes to update.
          Columns that don't exist on the model are skipped with a warning.
        :param always_set_last_update: force-set ``<attr>_last_update`` even if the new value is the same.
        :return: None
        """

        for k, v in updates.items():
            if not hasattr(self, k):
                warn("update_from_dict called on %r with non-existant attribute %r" % (self, k), RuntimeWarning)
                continue

            existing_value = getattr(self, k)
            setattr(self, k, v)

            if always_set_last_update or v != existing_value:
                last_update_column_name = f"{k}_last_update"
                if hasattr(self, last_update_column_name):
                    setattr(self, last_update_column_name, func.now())



import sqlalchemy
from sqlalchemy import Boolean
from sqlalchemy.orm import mapped_column, Mapped


class EnabledMixin:
    """Mixin that adds a boolean 'enabled' field that defaults to ``true``."""
    enabled: Mapped[bool] = mapped_column(Boolean, nullable=False, server_default=sqlalchemy.true())


class EnabledNoMixin:
    """Mixin that adds a boolean 'enabled' field that defaults to ``false``."""
    enabled: Mapped[bool] = mapped_column(Boolean, nullable=False, server_default=sqlalchemy.false())



from datetime import datetime

from sqlalchemy import func, DateTime
from sqlalchemy.orm import mapped_column, Mapped


class CreatedAtMixin:
    """SQLAlchemy mixin to add a `created_at` datetime field to a table."""
    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())


class UpdatedAtMixin:
    """
    SQLAlchemy mixin to add an `updated_at` datetime field to a table that’s set to `NOW()` every time the row is updated.
    Note that on Postgres this creates the column but update it only from Python. If you need database-level update, use
    a trigger.
    See https://stackoverflow.com/a/71072370/735926 for a possible solution.
    """
    # NOTE:
    #   server_default= tells the DB the default column value while default= sets the default value at the ORM level.
    #   onupdate= sets the value on record update at the ORM level.
    #   Contrary to what its name suggests, server_onupdates= doesn't tell anything to the DB. It only tells the ORM
    #   what to do when the record is updated, assuming the DB does that update by itself (e.g. via a trigger).
    #   See https://stackoverflow.com/a/33532154/735926.
    updated_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now(), onupdate=func.now())


class CreatedUpdatedAtMixin(CreatedAtMixin, UpdatedAtMixin):
    """SQLAlchemy mixin to add `created_at` and `updated_at` datetime fields to a table."""
